import { SpecialUnitCollection } from 'special-unit'

export default class DynamicCapture
{
  constructor() {
    // {text, specifiers, restrainers, modifier, modify_top_level, single_out}
    // this.specifiers = specifiers.toArray().map(this.unitize).toJson(Object.keys(specifiers))
    // this.captures = this.capture(
    //   text,
    //   this.specifiers,
    //   restrainers,
    //   modifier,
    //   modify_top_level,
    //   single_out
    // )
  }
  run({text, specifiers, restrainers, modifier, modify_top_level, single_out}) {
    this.specifiers = specifiers.toArray().map(this.unitize).toJson(Object.keys(specifiers))
    this.captures = this.capture(
      text,
      this.specifiers,
      restrainers,
      modifier,
      modify_top_level,
      single_out
    )
    return this
  }
  get() { return this.captures }
  capture(text, specifiers, restrainers, modifier, modify_top_level = true, single_out = true) {
    let variables = {}
    let remainder = []
    let keys = Object.keys(specifiers)
    keys.forEach(key => {
      variables[key] = []
    })
    while(text.length > 0){
      let matched = false
      for (let i=0; i<keys.length; i++){
        let specifier = specifiers[keys[i]]
        let length
        let match = this.matchSpecifierAtInitial(text, Object.create(specifier), restrainers, modifier, len => {
          length = len
        })
        if (match){
          matched = true
          // implement `single_out`
          if (single_out){
            if (Object.keys(match).length === 1){
              let k = Object.keys(match)[0]
              if (match[k] instanceof Object){
                match = match[k]
              }
            }
          }
          // apply modifier
          if (modify_top_level){
            let variable_name = keys[i]
            if (modifier)
            if (modifier[variable_name])
            match = modifier[variable_name](match)
          }
          variables[keys[i]].push(match)
          text = text.substr(length)
          break
        }
      }
      if (!matched){
        remainder.push(text[0])
        text = text.shift()
      }
    }
    variables['%%'] = remainder.join('')
    return variables
  }
  matchSpecifierAtInitial(text, specifier, restrainers, modifier, getReduced_length) {
    let len = text.length
    let matcher = specifier.shift()
    if (!text.initial(matcher.unit)) return null

    // remove top
    text = text.neckAndHeadOff(matcher.unit)

    let variable = {}
    let variable_name = null
    while (1){
      variable_name = this.stripVariable(specifier.shift().unit)
      matcher = specifier.shift()

      let head = text.getHead(matcher.unit)

      // check with restrainer
      if (restrainers[variable_name])
      if (head.match(restrainers[variable_name])) return null

      // apply modifier
      if (modifier)
      if (modifier[variable_name])
      head = modifier[variable_name](head)

      variable[variable_name] = head
      text = text.neckAndHeadOff(matcher.unit)

      if (specifier.length == 0) break
    }
    if (getReduced_length) getReduced_length(len - text.length)
    return variable
  }
  stripVariable(name) {
    name = name.split('')
    name.shift()
    name.shift()
    name.pop()
    return name.join('')
  }
  unitize(specifier) {
    const unit_specs = Object.freeze({
      variable: /\{\$[A-z\d_-]+\}/g
    })
    return new SpecialUnitCollection().setFromString(specifier, unit_specs).groupify()
  }
}